import yaml
from dynLoad import dynLoad

class Object:
    logFile = None
    
    def __init__(self, parent, type = None, name = None):
        self.parent = parent
        self.type = type
        self.name = name
        
        if not self.type:
            className = self.__class__.__name__
            self.type = className[:1].lower() + className[1:]
            
        if not self.name or self.parent.hasChild(self.name):
            if self.parent:
                self.name = self.type
                i = 1
                while self.parent.hasChild(self.name):
                    self.name = "%s%d" % (self.type,i)
                    i += 1
            else:
                self.name = "root"
        
        if self.parent:
            self.parent.children[self.name] = self
            if self.parent.parent:
                self.path = self.parent.path + "/" + self.name
            else:
                self.path = "/" + self.name
                
        else:
            self.path = "/"

        self.children = {}
        self.vars = {}
        self.initDepends = []
        self.destroyDepends = []
        self.subscribers = []
        self.childrenToRemove = []
        
        self.initialized = False
        self.destroyed = False
        self.active = True
        self.defaultVar("render",True)

        self.fileLoad(self.type)
        
        self.firstFrame = True
        
    def __del__(self):
        self.preDestroy()
        
    def isAlive(self):
        return self.initialized and self.active and not self.destroyed
        
    def fileLoad(self, type):
        try:
            fp = open("dat/objects/%s.yml" % type)
        except:
            return False
        
        data = yaml.load(fp)
        self.load(data)
        
        return True
        
    def load(self, data):
        for key in data.iterkeys():
            funcname = "load" + key[:1].upper() + key[1:]
            if hasattr(self,funcname) and callable(getattr(self,funcname)):
                getattr(self,funcname)(data[key])
            else:
                self.warn("unknown data type %s in file %s.yml" % (key,type))

    def loadChildren(self, data):
        for name, childData in data.iteritems():
            if childData.has_key('type'):
                type = childData['type']
            else:
                type = name
            child = (dynLoad(type) or Object)(self,type,name)
            child.load(childData)
            
    def loadVars(self, data):
        for key, val in data.iteritems():
            try:
                data[key] = eval(val)
            except:
                pass
        
        self.vars.update(data)
        
    def loadInitDepends(self, data):
        for path in data:
            self.addInitDepend(path)
            
    def loadDestroyDepends(self, data):
        for path in data:
            self.addDestroyDepend(path)
            
    def hasInitDepend(self, path):
        return path in self.initDepends
    
    def addInitDepend(self, path):
        if not self.hasInitDepend(path):
            self.initDepends.append(path)
            
    def removeInitDepend(self, path):
        if self.hasInitDepend(path):
            self.initDepends.remove(path)
            
    def hasDestroyDepend(self, path):
        return path in self.destroyDepends
    
    def addDestroyDepend(self, path):
        if not self.hasDestroyDepend(path):
            self.destroyDepends.append(path)
            
    def removeDestroyDepend(self, path):
        if self.hasDestroyDepend(path):
            self.destroyDepends.remove(path)
            
    def log(self, str):
        out = '%s: %s' % (self.path, str)
        print out
        
        if not Object.logFile:
            Object.logFile = open('log.txt','w')
        Object.logFile.write(out + '\n')
        Object.logFile.flush()
        
    def warn(self, str):
        self.log("WARNING: %s" % str)

    def error(self, str):
        self.log("*" * 50)
        self.log("ERROR: %s" % str)
        self.log("*" * 50)

    def preInit(self):
        if not self.initialized and not self.destroyed:
            self.initialized = True
            
            for path in self.initDepends:
                obj = self.getObject(path) or self.getFirstObjectOfName(path)
                if obj:
                    obj.preInit()
            
            self.init()
            
            for child in self.children.itervalues():
                child.preInit()
            
    def init(self):
        pass
            
    def preUpdate(self, time, dtime):
        if self.parent and self.parent.initialized and self.firstFrame:
            self.preInit()
        self.firstFrame = False
            
        if self.initialized and self.active and not self.destroyed:
            for childName in self.childrenToRemove:
                if self.children.has_key(childName):
                    child = self.children.pop(childName)
                    child.preDestroy()
            self.childrenToRemove = []
            
            self.update(time,dtime)
            
            for child in self.children.itervalues():
                child.preUpdate(time,dtime)
            
    def update(self, time, dtime):
        pass
            
    def preDestroy(self):
        if self.initialized and not self.destroyed:
            self.destroyed = True
            
            for path in self.destroyDepends:
                obj = self.getObject(path) or self.getFirstObjectOfName(path)
                if obj:
                    obj.preDestroy()
            
            self.destroy()
            
            for child in self.children.itervalues():
                child.preDestroy()


    def destroy(self):
        pass
    
    def hasVar(self, varName):
        return self.vars.has_key(varName)
    
    def getVar(self, varName):
        if self.hasVar(varName):
            return self.vars[varName]
        else:
            return None
        
    def setVar(self, varName, val):
        self.vars[varName] = val
        
    def defaultVar(self, varName, val):
        if not self.hasVar(varName):
            self.setVar(varName,val)
            
    def getRoot(self):
        if self.parent:
            return self.parent.getRoot()
        else:
            return self

    def hasChild(self, childName):
        return self.children.has_key(childName)
    
    def getChild(self, childName):
        if self.hasChild(childName):
            return self.children[childName]
        else:
            return None
        
    def getFirstObjectOfName(self, name, fromRoot = True):
        if fromRoot:
            return self.getRoot().getFirstObjectOfName(name,False)
        
        if self.name == name:
            return self
        
        for child in self.children.itervalues():
            obj = child.getFirstObjectOfName(name,False)
            if obj:
                return obj
            
        return None
        
    def getAllObjectsOfName(self, name, fromRoot = True):
        if fromRoot:
            return self.getRoot().getAllObjectsOfName(name,False)
        
        objs = []
        if self.name == name:
            objs.append(name)
            
        for child in self.children.itervalues():
            objs.extend(child.getAllObjectsOfName(name,False))
            
        return objs

    def getFirstObjectOfType(self, type, fromRoot = True):
        if fromRoot:
            return self.getRoot().getFirstObjectOfType(type,False)
        
        if self.type == type:
            return self
        
        for child in self.children.itervalues():
            obj = child.getFirstObjectOfType(type,False)
            if obj:
                return obj
            
        return None
        
    def getAllObjectsOfType(self, type, fromRoot = True):
        if fromRoot:
            return self.getRoot().getAllObjectsOfType(type,False)
        
        objs = []
        if self.type == type:
            objs.append(type)
            
        for child in self.children.itervalues():
            objs.extend(child.getAllObjectsOfType(type,False))
            
        return objs

    def getObject(self, path):
        if not path:
            return None
        
        #self.log("getObject %s" % path)
        
        if path[0] == '/':
            
            if len(path) > 1:
                #self.log("recursing to root")
                return self.getRoot().getObject(path[1:])
            
            else:
                #self.log("returning root")
                return self.getRoot()
        
        elif path.find('/') != -1:
            
            lead= path.split('/')[0]
            rest = path[len(lead)+1:]
            #self.log("lead = %s, rest = %s" % (lead, rest))
        
            if lead == '..':
                #self.log("recursing to parent")
                return self.getParent().getObject(rest)
            
            elif self.hasChild(lead):
                #self.log("recursing to child")
                return self.getChild(lead).getObject(rest)
            
            else:
                #self.log("not found")
                return None
            
        else:
            if path == '..':
                #self.log("returning parent")
                return self.getParent()
            
            elif self.hasChild(path):
                #self.log("returning child")
                return self.getChild(path)
            
            else:
                #self.log("not found")
                return None
            
    def subscribe(self, path):
        if not path.__class__ == str:
            path = path.path
        if not self.isSubscribed(path):
            self.subscribers.append(path)
            
    def unsubscribe(self, path):
        if self.isSubscribed(path):
            self.subscribers.remove(path)
            
    def isSubscribed(self, path):
        return path in self.subscribers
    
    def publish(self, cmd, msg):
        for path in self.subscribers:
            obj = self.getObject(path)
            obj.receiveMessage(cmd,msg)
            
    def receiveMessage(self, cmd, msg = None):
        if self.isAlive():
            funcname = 'message' + cmd[:1].upper() + cmd[1:]
            if hasattr(self,funcname) and callable(getattr(self,funcname)):
                getattr(self,funcname)(msg)

    def removeChild(self, childName):
        self.childrenToRemove.append(childName)
        
    def clearChildren(self):
        self.childrenToRemove.extend(self.children.keys())
        
    def deleteThis(self):
        if self.parent:
            self.parent.childrenToRemove.append(self.name)

