from object import Object
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import pygame, sys, math

from sceneNode import SceneNode
from circle import Circle
from rect import Rect
from triangle import Triangle

deg2rad = 0.017453292519943295
rad2deg = 57.295779513082323

class PygameGfxServer(Object):
    def __init__(self, parent):
        Object.__init__(self, parent, "pygameGfxServer", "gfxServer")
        
        self.defaultVar("renderMode",2)
        self.defaultVar("title", "PyGame Window")
        self.defaultVar("screenw",800)
        self.defaultVar("screenh",600)
        self.defaultVar("camerax",0)
        self.defaultVar("cameray",0)
        self.defaultVar("accelerate",True)
        self.defaultVar("textLineHeight",15)
        
        self.earlyObjects = []
        self.lateObjects = []
        self.lists = {}
        self.textures = {}
        self.recordingPath = None
        self.recordingList = None
        self.renderCount = 0
        
    def init(self):
        pygame.init()
        glutInit(sys.argv)
        
        screenw = self.getVar("screenw")
        screenh = self.getVar("screenh")
        renderMode = self.getVar("renderMode")
        
        self.screen = pygame.display.set_mode((screenw,screenh), pygame.OPENGL | pygame.DOUBLEBUF | pygame.HWSURFACE)
        pygame.display.set_caption(self.getVar("title"))
        
        glLoadIdentity()
        glMatrixMode(GL_PROJECTION)
        if renderMode == 3:
            gluPerspective(45.0,float(screenw)/float(screenh),1.0,500.0)
        else:
            gluOrtho2D(0,screenw,0,screenh)
        glMatrixMode(GL_MODELVIEW)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(0,0,0,0)
        
    def destroy(self):
        self.log("destroy")
        
        for id in self.textures.itervalues():
            glDeleteTextures([id])
        self.textures = {}
        
        for id in self.lists.itervalues():
            glDeleteLists(id,1)
        self.lists = {}
        
        self.log("pygame.quit")
        #pygame.quit()
        self.log("/pygame.quit")
        
        self.log("/destroy")
        
    def preRender(self):
        glMatrixMode(GL_MODELVIEW)
        glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT)
        glLoadIdentity()
        
        if self.getVar("renderMode") == 2:
            glTranslate(-self.getVar("camerax"),self.getVar("screenh")+self.getVar("cameray"),0)
            glScalef(1,-1,1)
            
        self.renderCount = 0
        
        for obj in self.earlyObjects:
            self.doRender(obj)
            
    def render(self, obj):
        if obj.isAlive() and obj.getVar("render") and obj not in self.earlyObjects and obj not in self.lateObjects:
            if issubclass(obj.__class__,SceneNode):
                self.doRender(obj)
            else:
                for child in obj.children.itervalues():
                    self.render(child)
                    
    def postRender(self):
        for obj in self.lateObjects:
            self.doRender(obj)
        
        glFlush()
        pygame.display.flip()

    def doRender(self, obj):
        glPushMatrix()
        glPushAttrib(GL_ALL_ATTRIB_BITS)
        
        glTranslate(0,0,0.001 * self.renderCount)
        self.renderCount += 1
            
        if self.lists.has_key(obj.path):
            glCallList(self.lists[obj.path])
            
        else:
            if self.getVar("accelerate") and not self.recordingPath and obj.getVar("static"):
                self.recordingPath = obj.path
                self.recordingList = glGenLists(1)
                glNewList(self.recordingList,GL_COMPILE_AND_EXECUTE)
            
            x,y,z = obj.getVar("pos")
            glTranslatef(x,y,z)
            
            y,p,r = [f * rad2deg for f in obj.getVar("rot")]
            glRotatef(y,0,1,0)
            glRotatef(p,1,0,0)
            glRotatef(r,0,0,1)
            
            x,y,z = obj.getVar("scale")
            glScalef(x,y,z)
            
            if obj.hasVar("color"):
                r,g,b = obj.getVar("color")
                glColor3f(r,g,b)
                
            if obj.hasVar("texture"):
                self.bindTexture(obj.getVar("texture"))
            else:
                self.unbindTexture()
                
            funcName = "render" + obj.__class__.__name__
            if hasattr(self,funcName) and callable(getattr(self,funcName)):
                getattr(self,funcName)(obj)
                
            for child in obj.children.itervalues():
                self.render(child)
                
            if self.getVar("accelerate") and obj.path == self.recordingPath:
                glEndList()
                self.lists[self.recordingPath] = self.recordingList
                self.recordingList = None
                self.recordingPath = None

        glPopAttrib()
        glPopMatrix()
        
    def renderCircle(self, obj):
        glBegin(GL_TRIANGLE_FAN)
        r,g,b = obj.getVar("centerColor")
        glColor3f(r,g,b)
        glTexCoord2f(0.5,0.5)
        glVertex3f(0,0,0)
        numSlices = obj.getVar("numSlices")
        r,g,b = obj.getVar("color")
        glColor3f(r,g,b)
        for i in range(numSlices+1):
            th = float(i) / float(numSlices) * math.pi * 2
            x = math.cos(th)
            y = math.sin(th)
            glTexCoord2f((x+1.0)/2.0,(y+1.0)/2.0)
            glVertex3f(x,y,0)
        glEnd()
        
    def renderRect(self, obj):
        glBegin(GL_QUADS)
        glTexCoord2f(0,0)
        glVertex3f(0,0,0)
        glTexCoord2f(0,1)
        glVertex3f(0,1,0)
        glTexCoord2f(1,1)
        glVertex3f(1,1,0)
        glTexCoord2f(1,0)
        glVertex3f(1,0,0)
        glEnd()
        
    def renderTriangle(self, obj):
        glBegin(GL_TRIANGLES)
        x,y,z = obj.getVar("v1")
        glTexCoord2f(x,y)
        glVertex3f(x,y,z)
        x,y,z = obj.getVar("v2")
        glTexCoord2f(x,y)
        glVertex3f(x,y,z)
        x,y,z = obj.getVar("v3")
        glTexCoord2f(x,y)
        glVertex3f(x,y,z)
        glEnd()
        
    def renderText(self, obj):
        text = obj.getVar("text")
        x,y,z = obj.getVar("pos")
        align = obj.getVar("align")
        glPushMatrix()
        cury = 0
        numLines = text.count('\n') + 1
        lineWidth = []
        for line in text.split('\n'):
            width = 0
            for i in line:
                width += glutBitmapWidth(GLUT_BITMAP_HELVETICA_12,ord(i))
            lineWidth.append(width)
        for i, line in enumerate(text.split('\n')):
            if align == 'left':
                startx = 0
            elif align == 'center':
                startx =  - lineWidth[i]/2
            else:
                startx =  - lineWidth[i]
            glRasterPos(startx,cury,z)
            cury += self.getVar("textLineHeight")
            for i in line:
                glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,ord(i))
        glPopMatrix()
            
    def loadTexture(self, filename):
        texture = pygame.image.load('dat/images/%s' % filename)
        w,h = texture.get_size()
        rgbTextureData = pygame.image.tostring(texture,"RGB",0)
        textureData = ""
        for i in range(0,len(rgbTextureData),3):
            r,g,b = [ord(j) for j in rgbTextureData[i:i+3]]
            if not r and not g and not b:
                a = 0
                #r = 255
                #g = 255
                #b = 0
            else:
                a = 255
            textureData += chr(r) + chr(g) + chr(b) + chr(a)
        textureId = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D,textureId)
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST)
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP)
        gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA,w,h,GL_RGBA,GL_UNSIGNED_BYTE,textureData)
        glBindTexture(GL_TEXTURE_2D,-1)
        return textureId

    def bindTexture(self, filename):
        if not self.textures.has_key(filename):
            self.textures[filename] = self.loadTexture(filename)
            
        id = self.textures[filename]
        glBindTexture(GL_TEXTURE_2D,id)
        
    def unbindTexture(self):
        glBindTexture(GL_TEXTURE_2D,-1)

